#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <string.h>
#include <ctype.h>

/* Standard defines needed by generated code */
#define RETURNTYPE void
#define PARAMDECL const uint32_t OPCODE
#define PARAMS OPCODE

/* Support code required by the actions */
const char *regnames[16] = {"R0","R1","R2","R3","R4","R5","R6","R7","R8","R9","R10","R11","R12","R13","R14","PC"};
const char *conditions[16] = {"EQ","NE","CS","CC","MI","PL","VS","VC","HI","LS","GE","LT","GT","LE","AL","NV"};
char buf[256];
char *bufpos;
bool incomment;

void comment(const char *str)
{
	if(!incomment)
	{
		while(bufpos-buf < 27)
			*bufpos++ = ' ';
		*bufpos++ = ';';
		incomment = true;
	}
	else
	{
		*bufpos++ = '.';
	}
	*bufpos++ = ' ';
	strcpy(bufpos,str);
	bufpos += strlen(bufpos);
}

void do_op(const char *op)
{
	strcpy(buf,op);
	bufpos = buf+strlen(buf);
	while(bufpos-buf < 8)
		*bufpos++ = ' ';
	incomment = false;
}

void do_reg(uint32_t r)
{
	if((bufpos != buf) && (bufpos[-1] != ' '))
		*bufpos++ = ',';
	strcpy(bufpos,regnames[r]);
	bufpos += strlen(bufpos);
}

void do_imm(uint32_t imm)
{
	if((bufpos != buf) && (bufpos[-1] != ' '))
		*bufpos++ = ',';
	if(imm > 255)
	{
		bufpos += sprintf(bufpos,"#&%04lX",imm);
		comment("=");
		bufpos += sprintf(bufpos,"%lu",(signed long) imm);
	}
	else if(imm<10)
		bufpos += sprintf(bufpos,"#%lu",imm);
	else
	{
		bufpos += sprintf(bufpos,"#&%02lX",imm);
		if(isprint(imm))
		{
			comment("=\"");
			*bufpos++ = imm;
			*bufpos++ = '"';
		}
		else
		{
			comment("=");
			bufpos += sprintf(bufpos,"%lu",(signed long) imm);
		}
	}
}

void do_hex(uint32_t h,uint32_t c)
{
	const char *hextab = "0123456789ABCDEF";
	if((bufpos != buf) && (bufpos[-1] != ' '))
		*bufpos++ = ',';
	*bufpos++ = '&';
	while(c--)
		*bufpos++ = hextab[(h>>(c<<2))&0xf];
}

void do_cond(uint32_t cond)
{
	/* fudgey! */
	char *c = buf;
	while(*c && (*c != ' '))
		c++;
	if(!c[0] || !c[1])
	{
		bufpos = c+2;
		*bufpos++ = ' ';
	}
	c[0] = conditions[cond][0];
	c[1] = conditions[cond][1];
}

void do_int(uint32_t imm)
{
	if((bufpos != buf) && (bufpos[-1] != ' '))
		*bufpos++ = ',';
	bufpos += sprintf(bufpos,"#%ld",(signed long) imm);
}

void do_reglist(uint32_t regs)
{
	int i;
	if((bufpos != buf) && (bufpos[-1] != ' '))
		*bufpos++ = ',';
	*bufpos++ = '{';
	for(i=0;i<16;i++)
	{
		if(regs & (1<<i))
		{
			if(bufpos[-1] != '{')
				*bufpos++ = ',';
			strcpy(bufpos,regnames[i]);
			bufpos += strlen(bufpos);
			if(regs & (1<<(i+1)))
			{
				while(regs & (1<<(i+1)))
					i++;
				bufpos += sprintf(bufpos,"-%s",regnames[i]);
			}
		}
	}
	*bufpos++ = '}';
}

#define op_cond_hex(o,cond,h,c) { do_op(o); do_cond(cond); do_hex(h,c); }
#define op_hex(o,h,c) { do_op(o); do_hex(h,c); }
#define op_reg(o,r1) { do_op(o); do_reg(r1); }
#define op_reglist(o,regs) { do_op(o); do_reglist(regs); }
#define op_reg_imm(o,r1,imm) { do_op(o); do_reg(r1); do_imm(imm); }
#define op_reg_hex(o,r1,h,c) { do_op(o); do_reg(r1); do_hex(h,c); }
#define op_reg_reg(o,r1,r2) { do_op(o); do_reg(r1); do_reg(r2); }
#define op_reg_reg_imm(o,r1,r2,imm) { do_op(o); do_reg(r1); do_reg(r2); do_imm(imm); }
#define op_reg_reg_int(o,r1,r2,imm) { do_op(o); do_reg(r1); do_reg(r2); do_int(imm); }
#define op_reg_reg_reg(o,r1,r2,r3) { do_op(o); do_reg(r1); do_reg(r2); do_reg(r3); }
#define op_reg_reglist(o,r1,wb,regs) { do_op(o); do_reg(r1); if(wb) *bufpos++ = '!'; do_reglist(regs); }
#define op_str(o,s) { do_op(o); strcpy(bufpos,s); bufpos += strlen(bufpos); }

void ldrstr_imm(const char *op,uint32_t r1,uint32_t r2,uint32_t imm)
{
	do_op(op);
	do_reg(r1);
	bufpos += sprintf(bufpos,",[%s,#%ld]",regnames[r2],(signed long) imm);
}

void ldrstr_reg(const char *op,uint32_t r1,uint32_t r2,uint32_t r3)
{
	do_op(op);
	do_reg(r1);
	bufpos += sprintf(bufpos,",[%s,%s]",regnames[r2],regnames[r3]);
}

uint32_t sextend(uint32_t i,uint32_t c)
{
	if(i & (1<<(c-1)))
		i |= 0xffffffff<<c;
	return i;
}

/* Main */
RETURNTYPE DECODER(PARAMDECL);

int main(int argc,char **argv)
{
	uint32_t i;
	for(i=0;i<65536;i++)
	{
		bufpos = buf;
		DECODER(i);
		if(bufpos != buf)
		{
			*bufpos = 0;
			puts(buf);
		}
	}
	return 0;
}
