/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/
#ifndef DECGEN_H
#define DECGEN_H

#include "common.h"

extern encodingset *compute_kin(const encodingset &e); /* Return set containing root kin of 'e' */

extern bool restrict_recurse(uint32_t fixed,uint32_t mask,const constraint **c,uint32_t num); /* Returns true if any opcode within fixed & mask satisfies all the constraints within c. May reorder c! */

extern bool contains_encoding(const opcode_subset *o,const encoding *e); /* Return true if o contains at least one opcode that matches e */

extern node *determine_best_node(const opcode_subset *o,const encodingset &e); /* Generate the best possible node to place in the decision tree for classifying the data (o,e) */

extern encodingset *split_on_constraint(const opcode_subset *o,const encodingset &e,const constraint *c); /* Split into two sets via constraint c */

extern encodingset *split_on_field_fast(const opcode_subset *o,const encodingset &e,uint32_t startbit,uint32_t bitcount); /* Faster version of split_on_field, only works if there are no opcode match/nomatch constraints */

extern encodingset *split_on_field(const opcode_subset *o,const encodingset &e,uint32_t startbit,uint32_t bitcount); /* Split into N sets via field */

extern infogain compute_gain(const encodingset *kin,const encodingset *es,uint32_t count); /* Compute information gain from splitting the kinset kin into the 'count' encoding sets 'es' */

extern void dedupe_tree(node *tree,bool v);

extern tokenset tokenise(const char *code);

extern bool compatible_tokensets(const tokenset &a,const tokenset &b);

extern void merge_tokensets(const vector<tokenset> &tokensets,tokenset &result,vector<map<string,string> > &arguments);

#endif
