/*
*Copyright(c)2014, Jeffrey Lee
*Allrightsreserved.
*
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met: 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
*/

#include "opcodesubset.h"
#include "encoding.h"
#include "constraint.h"

opcode_subset::opcode_subset()
{
	fixed = mask = 0;
}

opcode_subset::opcode_subset(const opcode_subset &copy)
{
	fixed = copy.fixed;
	mask = copy.mask;
	match = copy.match;
	nomatch = copy.nomatch;
}

opcode_subset::~opcode_subset()
{
}

void opcode_subset::constrain(uint32_t f,uint32_t m)
{
	/* Make sure we don't contradict ourselves */
	assert((fixed & m) == (f & mask));
	fixed |= f;
	mask |= m;
}

void opcode_subset::constrain(const constraint *c,bool shouldmatch)
{
	if(shouldmatch)
	{
		for(size_t i=0;i<match.size();i++)
			if(match[i] == c)
				return;
		match.push_back(c);
	}
	else
	{
		for(size_t i=0;i<nomatch.size();i++)
			if(nomatch[i] == c)
				return;
		nomatch.push_back(c);
	}
}

void opcode_subset::constrain(const encodingset &e)
{
	/* Iterate over encodings and for each one that matches build a map of which bits differ from the first matching opcode */
	uint32_t first=0;
	bool foundfirst = false;
	uint32_t opcode = 0;
	uint32_t varies = 0;
	size_t i=0;
	if(!e.size())
	{
		assert(fuzzy);
		return;
	}
	do {
		size_t j;
		/* Only consider opcodes that satisfy our constraints (the tree generator won't produce code that requires us to decode opcodes that don't match) */
		for(j=0;j<match.size();j++)
			if(!match[j]->eval(opcode | fixed))
				break;
		if(j == match.size())
		{
			/* Also skip opcodes that hit our nomatch list, because we'll never be asked to decode those */
			for(j=0;j<nomatch.size();j++)
				if(nomatch[j]->eval(opcode | fixed))
					break;
			if(j == nomatch.size())
			{
				for(j=0;j<e.size();j++,i++)
				{
					if(i==e.size())
						i = 0;
					if(e[i]->matches(opcode | fixed))
					{
						if(!foundfirst)
						{
							foundfirst = true;
							first = opcode;
						}
						else
							varies |= opcode ^ first;
						break;
					}
				}
			}
		}
		opcode++;
		while(opcode & mask)
			opcode = (opcode+mask)&~mask;
		/* TODO - Possible to skip the variable bits? */
	} while(opcode & opcodebits);
	assert(foundfirst);
	mask |= opcodebits & ~varies; /* Mask off the fixed bits */
	fixed |= first & mask; /* ... and their values */
	/* Now check if this new mask makes any constraints obsolete */
	for(i=0;i<e.size();i++)
		for(uint32_t j=0;j<e[i]->numconstraints;j++)
		{
			const constraint *c = e[i]->constraints[j];
			uint32_t field = c->getfields();
			if((field & mask) == field)
				constrain(c,c->eval(fixed));
		}
}

bool opcode_subset::matches(uint32_t opcode) const
{
	if((opcode & mask) != fixed)
		return false;
	for(size_t i=0;i<match.size();i++)
		if(!match[i]->eval(opcode))
			return false;
	return true;
}

bool opcode_subset::doesntmatch(uint32_t opcode) const
{
	for(size_t i=0;i<nomatch.size();i++)
		if(nomatch[i]->eval(opcode))
			return true;
	return false;
}
